from ._define import define, frozen
from ._field_specifiers import array, auto, field, static
from ._register import register_fieldz

__all__ = ["array", "auto", "define", "field", "frozen", "register_fieldz", "static"]
